__author__ = 'Dominik Zuercher, Joerg Herbel, Silvan Fischbacher'
__email__ = 'dominik.zuercher@phys.ethz.ch, joerg.herbel@phys.ethz.ch, silvanf@phys.ethz.ch'  # noqa
__credits__ = 'ETH Zurich, Institute for Particle Physics and Astrophysics'

__version__ = "1.10.1"
